import dataclasses
import re
import typing
from collections import Counter
from dataclasses import dataclass
from typing import Type

from rdflib import Graph
from rdflib import RDFS, RDF, OWL
from rdflib import URIRef, Literal, BNode
from rdflib.term import Node

from verbalizer.nlp import ParaphraseLanguageModel
from verbalizer.patterns import Pattern
from verbalizer.vocabulary import Vocabulary

_RE_COMBINE_WHITESPACE = re.compile(r"\s+")

default_patterns = []


class VerbalizationEdge:

    def __init__(self, relationship, node: 'VerbalizationNode'):
        self.relationship = relationship
        self.node = node
        self._display = None

    @property
    def display(self):
        """
        Get the edge display (relationship label)
        """
        return self._display

    @display.setter
    def display(self, value):
        """
        Set the edge display. Can be set only if not set already. Otherwise the value is ignored.
        :param value: The value.
        """
        if not self._display:
            self._display = value

    def __repr__(self):
        return self.display

    def verbalize(self) -> str:
        """
        Traverse the edge to the next node to perform verbalization. This is a recursive operation.
        :return The verbalized text.
        """
        display = self.display

        if not display:
            display = ''

        if display.startswith('#'):
            display = ''

        node = self.node.verbalize().strip()
        if node:
            next_text = f' {node}'
        else:
            next_text = ''

        return f'{display}{next_text}'


class VerbalizationNode:
    def __init__(self, concept, parent_path=None):
        self.concept = concept
        if parent_path is None:
            self._true_path = [(concept, None)]
        else:
            self._true_path = parent_path + [(concept, None)]
        self.references: list[VerbalizationEdge] = []
        self._display = None

    def add_edge(self, edge: VerbalizationEdge):
        """
        Add a new edge to the node.
        :param edge: New edge.
        """
        self.references.append(edge)

    def get_path(self):
        """
        Get the full path including the current node.
        :return: List of tuple (URIRef, URIRef)
        """
        return list(self._true_path)

    def get_parent_path(self):
        """
        Get the full path excluding the current node.
        :return: List of tuple (URIRef, URIRef)
        """
        return list(self._true_path[:-1])

    def get_next_node(self, relationship, concept):
        for reference in self.references:
            if reference.relationship == relationship and reference.node.concept == concept:
                return reference.node
        return None

    @property
    def display(self):
        """
        Get the node display (concept label)
        """
        return self._display

    @display.setter
    def display(self, value):
        """
        Set the node display. Can be set only if not set already. Otherwise the value is ignored.
        :param value: The value.
        """
        if self._display is None:
            self._display = value

    def __repr__(self):
        return self.display

    def verbalize(self) -> str:
        """
        Traverse the node to the next node via the edges to perform verbalization. This is a recursive operation.
        :return The verbalized text.
        """
        # Resolve sentences from edges
        sentences = [edge.verbalize().strip() for edge in self.references]

        display = self.display

        # For Blank Nodes, remove the display.
        if isinstance(self.concept, BNode):
            display = ''

        # Group multiple sentences together using `and`.
        if len(sentences) >= 2:
            next_text = f' ({", and ".join(sentences)})'
        elif len(sentences) == 1:
            next_text = f' {sentences[0]}'
        else:
            next_text = ''

        # Add indefinite articles
        if not isinstance(display, str) or not display:
            indefinite_article = ''
        elif display[0] in {'a', 'e', 'i', 'o', 'u'}:
            indefinite_article = 'an '
        else:
            indefinite_article = 'a '

        return f'{indefinite_article}{display}{next_text}'


@dataclass
class VerbalizerModelUsageConfig:
    """
    Object used to configure when and how LLM shall be used.
    """
    min_patterns_evaluated: int = 0
    min_statements: int = 1
    extra_context: str = ""


@dataclass
class VerbalizerInstanceStats:
    """
    Object used to collect statistics.
    """
    patterns_evaluated = 0
    statements = 0
    relationship_counter: Counter = dataclasses.field(default_factory=Counter)
    concepts: set = dataclasses.field(default_factory=set)


class Verbalizer:
    prefix = 'https://zaitoun.dev/onto/'

    def __init__(
            self,
            vocabulary: Vocabulary,
            patterns: list[Type[Pattern]] = None,
            language_model: ParaphraseLanguageModel = None,
            usage_config: VerbalizerModelUsageConfig = None
    ):
        self.graph = vocabulary.graph
        self.vocab = vocabulary
        self.llm = language_model
        self.llm_config = usage_config or VerbalizerModelUsageConfig(0,2,"")
        self.patterns = [pattern(self.graph, self, vocabulary) for pattern in patterns or default_patterns]

    def verbalize(self, starting_concept: typing.Union[str, URIRef]) -> (str, str, str, VerbalizerInstanceStats):
        """
        Returns the Turtle fragment, CNL statement, LLM verbalized textual description, and stats.
        :param starting_concept: The URI of the concept to verbalize.
        :return: (fragment, CNL text, LLM text, stats)
        """
        sentences = set()
        triples = []
        stats = VerbalizerInstanceStats()

        if isinstance(starting_concept, str):
            starting_concept = URIRef(starting_concept)

        node = VerbalizationNode(starting_concept)
        self._verbalize_as_text_from(node, self.vocab, triples, stats)

        for ref in node.references:
            sentences.add(_RE_COMBINE_WHITESPACE.sub(" ", f'{node.display} {ref.verbalize().strip()}.').strip())

        text = '\n'.join(sorted(sentences))
        onto_fragment: str = self.generate_fragment(triples)

        # update stats
        for triple in triples:
            subject, predicate, obj = triple
            stats.relationship_counter[predicate] += 1

            if isinstance(subject, URIRef):
                stats.concepts.add(subject)

            if isinstance(obj, URIRef):
                stats.concepts.add(obj)

        stats.statements = len(sentences)

        # convert the pseudo text into proper English.
        use_llm = bool(self.llm and self._check_llm_usage_policy(stats))
        llm_text = None
        if use_llm:
            llm_text = self.llm.pseudo_to_text(text, extra=self.llm_config.extra_context)

        return onto_fragment, text, llm_text, stats

    def _verbalize_as_text_from(self,
                                node: VerbalizationNode,
                                vocab: Vocabulary,
                                triple_collector: list[tuple[Node, Node, Node]],
                                stats: VerbalizerInstanceStats):
        """
        Receives an input node which holds information about some concept. The node is then expanded into a tree-like
        graph by making queries against the knowledge base. Each time the function is called, the neighbours of the
        most recent nodes are queried and are used to expand the "graph". This results in an expanded node object which
        can then be used for verbalization by walking over all the different paths created.

        :param node: The starting node to verbalize from
        :param vocab: Used as a lookup vocabulary
        :param triple_collector: Used to collect triples
        :param stats: used to collect statistics
        :return: None
        """
        query = self.next_step_query_builder(node)
        results = self.graph.query(query)

        results_normalized = False

        # check patterns
        for pattern in self.patterns:
            if not pattern.check(results):
                continue

            results = pattern.normalize(node, triple_collector)
            results_normalized = True
            stats.patterns_evaluated += 1
            break

        # set node display
        node.display = vocab.get_class_label(node.concept)

        for result in results:
            relation, obj_2 = result

            relation_display = vocab.get_relationship_label(relation)

            if relation_display == Vocabulary.IGNORE_VALUE:
                if self.vocab.should_keep(relation):
                    triple_collector.append((node.concept, relation, obj_2))
                continue

            if not results_normalized:
                # Continue to expand the graph
                next_node = VerbalizationNode(obj_2, parent_path=node.get_parent_path() + [(node.concept, relation)])
                edge = VerbalizationEdge(relation, next_node)
                node.add_edge(edge)
                edge.display = relation_display

                # collect triple
                triple_collector.append((node.concept, relation, next_node.concept))
            else:
                # we already built this part of the graph so we just need to fetch it out.
                next_node = node.get_next_node(relation, obj_2)

            if isinstance(obj_2, URIRef):
                obj_display_2 = vocab.get_class_label(obj_2)
            elif isinstance(obj_2, Literal):
                obj_display_2 = obj_2.toPython()
            elif isinstance(obj_2, BNode):
                self._verbalize_as_text_from(next_node, vocab, triple_collector, stats)
                continue
            else:
                obj_display_2 = None

            next_node.display = obj_display_2

    @classmethod
    def next_step_query_builder(cls, node: VerbalizationNode) -> str:
        """
        Given a list of RDF Nodes, where each node is somehow connected to the next node in the list,
        return a SPARQL query expression to get all the relationships of the last node.
        This is needed in order to traverse over Blank Nodes.

        Consider the following graph:
        A -[rel]-> B
        B -[rel]-> C
        B -[rel]-> D

        If the input is:
        A-[rel]->B (The path)
        Then the output is query that would return:
        B -[rel]-> C
        B -[rel]-> D
        """

        trail = node.get_path()
        expressions = []
        filters = []
        for i, (t, p) in enumerate(trail):
            is_last = i == len(trail) - 1
            ref = cls._get_reference_expression(t, index=i)
            if is_last:
                exp = f'{ref} ?p ?o'
            else:
                next_t, _ = trail[i + 1]
                next_ref = cls._get_reference_expression(next_t, index=i + 1)
                exp = f'{ref} <{p.toPython()}> {next_ref}'

            if isinstance(t, BNode):
                filters.append(f"filter(str(?o{i}) = '{t.toPython()}')")

            expressions.append(exp)

        query = f"SELECT ?p ?o WHERE {{ {' . '.join(expressions)} . {' '.join(filters)} }}"
        return query

    def generate_fragment(self, triples: list[tuple[Node, URIRef, Node]], add_labels=False) -> str:
        """
        Return a string that represents the triples as an ontology fragment in turtle format.
        """

        def display_to_uri(display: str) -> URIRef:
            if display.startswith('http'):
                return URIRef(display)
            identifier = re.sub('[^a-zA-Z0-9 \n\.]', ' ', display).lower().replace(' ', '_')
            return URIRef(self.prefix + identifier)

        g = Graph()
        g.bind(prefix='', namespace=self.prefix)
        for triple in triples:
            subject, predicate, obj = triple

            subject_node = subject
            predicate_node = predicate
            object_node = obj

            if isinstance(subject, URIRef):
                subject_vocab_rep = self.vocab.get_class_label(subject, default=subject.toPython())
                if subject_vocab_rep == Vocabulary.IGNORE_VALUE:
                    continue
                subject_node = display_to_uri(subject_vocab_rep)

            if isinstance(obj, URIRef) and not obj.toPython().startswith(str(OWL)):
                object_vocab_rep = self.vocab.get_class_label(obj, default=obj.toPython())
                if object_vocab_rep == Vocabulary.IGNORE_VALUE:
                    continue
                object_node = display_to_uri(object_vocab_rep)

            if not self._starts_with_one_of(predicate_node.toPython(), [OWL, RDF, RDFS]):
                label = self.vocab.get_relationship_label(predicate)
                if label == Vocabulary.IGNORE_VALUE:
                    continue
                predicate_node = display_to_uri(label)

            g.add((subject_node, predicate_node, object_node))

            # If subject or object have labels - add them as rdfs:label
            if add_labels:
                if isinstance(subject, URIRef) and self.vocab.get_class_label(subject):
                    g.add((subject_node, RDFS.label, Literal(self.vocab.get_class_label(subject))))

                if isinstance(obj, URIRef) and \
                        self.vocab.get_class_label(obj) and \
                        not obj.toPython().startswith(str(OWL)):
                    g.add((object_node, RDFS.label, Literal(self.vocab.get_class_label(obj))))

        fragment = g.serialize(format='turtle')
        return '\n'.join(fragment.split("\n")[1:])

    @staticmethod
    def _starts_with_one_of(val: str, items: list):
        """
        Helper function to check if list contains an element that starts with the provided value.
        :param val: The value to check.
        :param items: The list
        :return: True if at least one item begins with the provided string.
        """
        return any([val.startswith(str(e)) for e in items])

    def _check_llm_usage_policy(self, stats: VerbalizerInstanceStats) -> bool:
        """
        Helper function used to check if LLM should be invoked.
        :param stats: Stats collected.
        :return: True if met.
        """
        return all([
            stats.patterns_evaluated >= self.llm_config.min_patterns_evaluated,
            stats.statements >= self.llm_config.min_statements
        ])

    @classmethod
    def _get_reference_expression(cls, t: Node, index: int):
        """
        Util function, used to convert Node t to a queryable expression.
        """
        if isinstance(t, URIRef):
            return f'<{t.toPython()}>'
        if isinstance(t, BNode):
            return f'?o{index}'

        return None



from rdflib import URIRef

from verbalizer.patterns import Pattern
from verbalizer.verbalizer import VerbalizationNode, VerbalizationEdge, default_patterns


class OwlRestrictionPattern(Pattern):
    """
    Pattern that handles restrictions
    """

    def check(self, results) -> bool:
        expected = {
            'http://www.w3.org/2002/07/owl#onProperty',
            'http://www.w3.org/2002/07/owl#someValuesFrom',
            'http://www.w3.org/2002/07/owl#allValuesFrom',
            'http://www.w3.org/2002/07/owl#hasValue',
            'http://www.w3.org/2002/07/owl#cardinality',
            'http://www.w3.org/2002/07/owl#minCardinality',
            'http://www.w3.org/2002/07/owl#maxCardinality',
            'http://www.w3.org/2002/07/owl#qualifiedCardinality',
            'http://www.w3.org/2002/07/owl#minQualifiedCardinality',
            'http://www.w3.org/2002/07/owl#maxQualifiedCardinality',
            'http://www.w3.org/2002/07/owl#onClass'
        }
        actual = {relation.toPython() for (relation, obj) in results}

        return len(expected.intersection(actual)) >= 2

    def normalize(self, node: VerbalizationNode, triple_collector):
        query = self.verbalizer.next_step_query_builder(node)
        results = self._graph.query(query)

        next_node = None
        quantifier_relation = None
        property_relation = None
        literal_value = None
        on_class = None

        for (relation, obj) in results:

            if self.vocab.should_ignore(relation.toPython()):
                triple_collector.append((node.concept, relation, obj))
                continue

            if relation.toPython() == 'http://www.w3.org/2002/07/owl#onProperty':
                property_relation = obj

            if relation.toPython() in {'http://www.w3.org/2002/07/owl#someValuesFrom',
                                       'http://www.w3.org/2002/07/owl#allValuesFrom',
                                       'http://www.w3.org/2002/07/owl#hasValue'}:
                quantifier_relation = relation
                next_node = VerbalizationNode(concept=obj,
                                              parent_path=node.get_parent_path() + [(node.concept, relation)])

            if relation.toPython() in {'http://www.w3.org/2002/07/owl#cardinality',
                                       'http://www.w3.org/2002/07/owl#minCardinality',
                                       'http://www.w3.org/2002/07/owl#maxCardinality',
                                       'http://www.w3.org/2002/07/owl#qualifiedCardinality',
                                       'http://www.w3.org/2002/07/owl#minQualifiedCardinality',
                                       'http://www.w3.org/2002/07/owl#maxQualifiedCardinality'
                                       }:
                quantifier_relation = relation
                literal_value = obj

                # initialize next_node only if it hasn't be initialized yet.
                if next_node is None:
                    next_node = VerbalizationNode(
                        concept='',
                        parent_path=node.get_parent_path() + [(node.concept, relation)]
                    )
                    next_node.display = ''

            if relation.toPython() in {
                'http://www.w3.org/2002/07/owl#onClass'
            }:
                on_class = obj
                next_node = VerbalizationNode(
                    concept=obj,
                    parent_path=node.get_parent_path() + [(node.concept, relation)]
                )
                next_node.display = ''

            triple_collector.append((node.concept, relation, obj))

        edge = VerbalizationEdge(
            relationship=URIRef(quantifier_relation.toPython() + property_relation.toPython()),
            node=next_node
        )

        quantifier_relation_label = self.vocab.get_relationship_label(quantifier_relation)
        property_relation_label = self.vocab.get_class_label(property_relation)

        if quantifier_relation.toPython().endswith('someValuesFrom'):
            edge.display = f'at least {property_relation_label} some'
        elif quantifier_relation.toPython().endswith('allValuesFrom'):
            edge.display = f'only {property_relation_label}'
        elif quantifier_relation.toPython().endswith('hasValue'):
            edge.display = f'must {property_relation_label}'
        elif quantifier_relation.toPython().lower().endswith('cardinality'):
            edge.display = self._handle_cardinality(quantifier_relation, property_relation, literal_value, on_class)
        else:
            edge.display = f'{property_relation_label} {quantifier_relation_label}'

        node.add_edge(edge)

        return [(reference.relationship, reference.node.concept) for reference in node.references]

    def _handle_cardinality(self, quantifier_relation, property_relation, obj_literal, on_class) -> str:
        property_relation_label = self.vocab.get_class_label(property_relation)
        literal_value = obj_literal.toPython()
        on_class_label = ' '

        relation_plural_s = 's' if literal_value > 1 and not property_relation_label.endswith('s') else ''

        if property_relation_label.startswith('has'):
            property_relation_label = property_relation_label.replace('has ', '')

        if on_class:
            on_class_label = f' {self.vocab.get_class_label(on_class)} '

        if quantifier_relation.endswith('cardinality') and literal_value == 0:
            return f'has zero {on_class_label} {property_relation_label}s'
        elif quantifier_relation.endswith('cardinality') or quantifier_relation.endswith('qualifiedCardinality'):
            return f'has exactly {literal_value}{on_class_label}{property_relation_label}{relation_plural_s}'
        elif quantifier_relation.endswith('minCardinality') or quantifier_relation.endswith('minQualifiedCardinality'):
            return f'has at least {literal_value}{on_class_label}{property_relation_label}{relation_plural_s}'
        elif quantifier_relation.endswith('maxCardinality') or quantifier_relation.endswith('maxQualifiedCardinality'):
            return f'has at most {literal_value}{on_class_label}{property_relation_label}{relation_plural_s}'

default_patterns.append(OwlRestrictionPattern)


from rdflib import URIRef

from verbalizer.patterns import Pattern
from verbalizer.verbalizer import VerbalizationNode, VerbalizationEdge, default_patterns
from verbalizer.vocabulary import Vocabulary


class OwlDisjointWith(Pattern):
    """
    Pattern that handles multiple class disjoints in OWL. It works by combining disjoints of the same object into a
    "collection". This results in having one statement for all disjoints.
    """
    disjoint_relation = 'http://www.w3.org/2002/07/owl#disjointWith'

    def check(self, results) -> bool:
        return self.disjoint_relation in {relation.toPython() for (relation, obj) in results}

    def normalize(self, node: VerbalizationNode, triple_collector):

        # Separate results into two groups: 1) related to disjointness, 2) all other
        query = self.verbalizer.next_step_query_builder(node)
        query_results = self._graph.query(query)

        # create intermediate node
        intermediate_node = VerbalizationNode(
            '_', parent_path=node.get_parent_path() + [(node.concept, URIRef(self.disjoint_relation))]
        )
        intermediate_node.display = ''
        intermediate_edge = VerbalizationEdge(URIRef(self.disjoint_relation), intermediate_node)
        intermediate_edge.display = self.vocab.get_relationship_label(self.disjoint_relation)
        node.add_edge(intermediate_edge)
        for (relation, obj) in query_results:
            if relation != URIRef(self.disjoint_relation):
                continue
            next_node = VerbalizationNode(obj, parent_path=node.get_parent_path() + [(node.concept, relation)])
            next_node.display = self.vocab.get_class_label(obj)
            edge = VerbalizationEdge(URIRef('http://www.w3.org/1999/02/22-rdf-syntax-ns#collection'), next_node)
            edge.display = '#collection'
            intermediate_node.add_edge(edge)
            triple_collector.append((node.concept, relation, obj))

        # take care of non-disjoint relationships. This basically works the same way as without a pattern.
        for (relation, obj) in query_results:
            if relation == URIRef(self.disjoint_relation):
                continue
            relation_display = self.vocab.get_relationship_label(relation)

            if relation_display == Vocabulary.IGNORE_VALUE:
                if self.vocab.should_keep(relation):
                    triple_collector.append((node.concept, relation, obj))
                continue

            next_node = VerbalizationNode(obj, parent_path=node.get_parent_path() + [(node.concept, relation)])
            edge = VerbalizationEdge(relation, next_node)
            node.add_edge(edge)
            edge.display = relation_display
            triple_collector.append((node.concept, relation, obj))

        return [(reference.relationship, reference.node.concept) for reference in node.references]

default_patterns.append(OwlDisjointWith)


import logging
from abc import ABC, abstractmethod
from typing import Optional

try:
    from openai import OpenAI

    logging.getLogger("openai").setLevel(logging.ERROR)
    logging.getLogger("httpx").setLevel(logging.ERROR)
except ModuleNotFoundError as err:
    OpenAI = None


class ParaphraseLanguageModel(ABC):

    @abstractmethod
    def pseudo_to_text(self, pseudo_text: str, extra: str = None) -> str:
        """
        Given a pseudo text or controlled natural language, return a rephrased version of that same text.
        :param pseudo_text: The CNL set of statements,
        :param extra: Additional context to include as part of the prompt.
        :return: Paraphrased text.
        """
        return pseudo_text

    @property
    def cost(self) -> float:
        """
        The usage cost so far of the model.
        """
        return 0.0

    @property
    def name(self) -> str:
        """
        The name of the model used.
        """
        return 'Unknown'


def get_messages(pseudo_text: str, extra_context: Optional[str] = None):
    """
    Get messages (system, assistant, user)
    """
    extra = extra_context if extra_context else ""
    return [
        {
            "role": "system",
            "content": f"You are an extremely specific data expert capable of converting pseudo English sentences "
                       f"into a meaningful and casual paragraph without losing information. Avoid repeating "
                       f"information. Spell out everything don't be lazy! {extra}"
        },
        {
            "role": "user", "content":
            """
            x relation 1 y
            z relation 2 x
            m relation 3 n
            """
        },
        {
            "role": "assistant", "content":
            """
            X has this relation 1 with Y, Z shares a relation 2 with X.
            Moreover, m has relation 3 with n.
            """
        },
        {"role": "user", "content": "X is same as something that intersection of something that something that has at "
                                    "least 3 N and M."},
        {"role": "assistant", "content": "X is the same as M which has at least three N"},
        {"role": "user", "content":
            """
            X is a type of at least has Y relation some a M.
            X is a type of at least has Y relation some a N.
            X is a type of only has Y relation any of (a M and a N)
            """
         },
        {"role": "assistant", "content": "X has Y relation with M and N and nothing else."},
        {"role": "user", "content": pseudo_text},
    ]


class ChatGptModelParaphrase(ParaphraseLanguageModel):
    """
    OpenAI wrapper implementation.
    """

    models = {
        "gpt-4o": {
            "input": 0.005,
            "output": 0.015
        },
        "gpt-4-0125-preview": {
            "input": 0.01,
            "output": 0.03
        },
        "gpt-4-1106-preview": {
            "input": 0.01,
            "output": 0.03
        },
        "gpt-4-1106-vision-preview": {
            "input": 0.01,
            "output": 0.03
        },
        "gpt-4": {
            "input": 0.03,
            "output": 0.06
        },
        "gpt-4-0613": {
            "input": 0.03,
            "output": 0.06
        },
        "gpt-4-32k": {
            "input": 0.06,
            "output": 0.12
        },
        "gpt-3.5-turbo-0125": {
            "input": 0.0005,
            "output": 0.0015
        },
        "gpt-3.5-turbo-instruct": {
            "input": 0.0015,
            "output": 0.0020
        },
        "gpt-3.5-turbo-16k-0613": {
            "input": 0.0030,
            "output": 0.0040
        },
        "gpt-3.5-turbo-1106": {
            "input": 0.0010,
            "output": 0.0020
        },
        "gpt-3.5-turbo-0613": {
            "input": 0.0015,
            "output": 0.0020
        },
        "gpt-3.5-turbo-0301": {
            "input": 0.0015,
            "output": 0.0020
        }
    }

    def __init__(self, api_key: str, model: str = 'gpt-3.5-turbo-0613', temperature=0.5):
        if not OpenAI:
            raise ModuleNotFoundError("OpenAI is not installed. Please install it with `pip install openai`")

        self.model = model
        self.temperature = temperature
        self.client = OpenAI(api_key=api_key)
        self._in_token_usage = 0
        self._out_token_usage = 0

    def pseudo_to_text(self, pseudo_text: str, extra: str = None) -> str:
        response = self.client.chat.completions.create(
            model=self.model,
            messages=get_messages(pseudo_text, extra),
            temperature=self.temperature
        )
        self._in_token_usage += response.usage.prompt_tokens
        self._out_token_usage += response.usage.completion_tokens
        return response.choices[0].message.content

    @property
    def cost(self) -> float:
        model_pricing = self.models.get(self.model) or {'input': 0.0, 'output': 0.0}

        in_tokens = self._in_token_usage / 1000
        out_tokens = self._out_token_usage / 1000

        return in_tokens * model_pricing['input'] + out_tokens * model_pricing["output"]

    @property
    def name(self) -> str:
        return self.model



from verbalizer.patterns import Pattern
from rdflib import URIRef

from verbalizer.verbalizer import VerbalizationNode, VerbalizationEdge, default_patterns


class OwlFirstRestPattern(Pattern):
    """
    Pattern that handles ordered lists (first-rest) in OWL.
    """

    def check(self, results) -> bool:
        expected = {'http://www.w3.org/1999/02/22-rdf-syntax-ns#first',
                    'http://www.w3.org/1999/02/22-rdf-syntax-ns#rest'}
        return {relation.toPython() for (relation, obj) in results} == expected

    def normalize(self, node: VerbalizationNode, triple_collector):
        current = node
        while current.concept != URIRef('http://www.w3.org/1999/02/22-rdf-syntax-ns#nil'):
            query = self.verbalizer.next_step_query_builder(current)
            results = self._graph.query(query)
            rest_node = None
            for (relation, obj) in results:
                next_node = VerbalizationNode(
                    obj, parent_path=current.get_parent_path() + [(current.concept, relation)]
                )
                if relation == URIRef('http://www.w3.org/1999/02/22-rdf-syntax-ns#first'):
                    # first
                    edge = VerbalizationEdge(URIRef('http://www.w3.org/1999/02/22-rdf-syntax-ns#collection'), next_node)
                    node.add_edge(edge)
                    edge.display = '#collection'
                else:
                    # rest
                    rest_node = next_node
                triple_collector.append((current.concept, relation, obj))

            current = rest_node

        return [(reference.relationship, reference.node.concept) for reference in node.references]

default_patterns.append(OwlFirstRestPattern)


import random


class Sampler:

    def __init__(self, *, sample_percentage: float = None, sample_n=None, seed=None):
        """
        Can only specify one of these parameters
        :param sample_percentage: The percentage of each list to sample (0 < sample_percentage <= 1)
        :param sample_n: The number of items to sample from each list (sample_n > 0)
        """
        if (sample_percentage is None) == (sample_n is None):
            raise ValueError("Exactly one of sample_percentage or sample_n must be specified")

        if sample_percentage is not None:
            if not (0 < sample_percentage <= 1):
                raise ValueError("sample_percentage must be between 0 and 1")

        if sample_n is not None:
            if sample_n <= 0:
                raise ValueError("sample_n must be greater than 0")

        self.sample_percentage = sample_percentage
        self.sample_n = sample_n
        self.random_gen = random.Random(seed)

    def get_sample(self, items: dict[str, list]) -> dict[str, list]:
        """
        Given a dictionary where the key is list id, and the value is a list,
        return a new dictionary with a sampled list per key based on the sampling configuration.
        """

        sampled_dict = {}

        for key, value_list in items.items():
            if self.sample_percentage is not None:
                n_samples = int(len(value_list) * self.sample_percentage)
            else:
                n_samples = min(self.sample_n, len(value_list))

            sampled_dict[key] = self.random_gen.sample(value_list, n_samples)

        return sampled_dict



import re
import logging
from rdflib import URIRef, Graph
from tqdm import tqdm

logger = logging.getLogger(__name__)


class Vocabulary:
    """
    Vocabulary class, used as a URI to term lookup.
    """

    IGNORE_VALUE = object()

    RELATIONSHIP_QUERY = """
    SELECT DISTINCT ?p ?relation
    WHERE {
        ?o1 ?p ?o2 .
        OPTIONAL {
            ?p rdfs:label ?pLabel
        }
        BIND (
          COALESCE(
            ?pLabel,
            ?p
          ) AS ?relation
        )
    }
    """

    OBJECTS_QUERY = """
    SELECT DISTINCT ?o1 ?o1Label
    WHERE {
        {
            SELECT ?o1 ?o1Label WHERE {
                ?o1 ?p ?o2 .
                ?o1 rdfs:label ?o1Label
            }
        }
        UNION
        {
            SELECT ?o1 ?o1Label WHERE {
                ?o1 a owl:Class .
                OPTIONAL {
                    ?o1 rdfs:label ?optionalLabel
                }
                BIND (
                  COALESCE(
                    ?optionalLabel,
                    ?o1
                  ) AS ?o1Label
                )
            }
        }    
    }
    """

    def __init__(self, graph: Graph, ignore: set[str] = None, guard: set[str] = None, rephrased: dict[str, str] = None):
        """
        :param graph: The ontology.
        :param ignore: URIs to ignore.
        :param guard: URIs to keep in the fragment even if they are in the ignore list.
        :param rephrased: URIs to rephrase/rename
        """
        logger.info('Initializing vocabulary')
        self.graph = graph
        self.relationship_labels = self._get_ontology_relationship_labels()
        self.object_labels = self._get_ontology_object_labels()
        self._load_imports()
        self.rephrased = rephrased or dict()
        self._ignore_list = ignore or {}
        self._guard_list = guard or {
            'http://www.w3.org/1999/02/22-rdf-syntax-ns#type',
            'http://www.w3.org/2000/01/rdf-schema#label',
            'http://www.w3.org/2002/07/owl#onDatatype'
        }

    def should_ignore(self, uri: str | URIRef) -> bool:
        """
        Check if URI should be ignored or not.
        """
        if isinstance(uri, URIRef):
            uri = uri.toPython()
        return uri in self._ignore_list

    def should_keep(self, uri: str | URIRef) -> bool:
        """
        Check if URI should be kept in the fragment or not.
        """
        if isinstance(uri, URIRef):
            uri = uri.toPython()
        return uri in self._guard_list

    def get_relationship_label(self, val, default=None) -> str:
        """
        Get relationship label.
        :param val: The URI to lookup with.
        :param default: Default value if not found.
        :return: string.
        """
        return self._util_lookup(self.relationship_labels, val) or default

    def get_class_label(self, val, default=None) -> str:
        """
        Get class label.
        :param val: The URI to lookup with.
        :param default: Default value if not found.
        :return: string.
        """
        return self._util_lookup(self.object_labels, val) or default

    def _get_ontology_relationship_labels(self) -> dict[str, str]:
        """
        Returns a IRI (URI) to label dictionary.
        """
        results = self.graph.query(self.RELATIONSHIP_QUERY)
        ontology_relations = {}

        for result in tqdm(results, desc='Loading Ontology Relationship Labels'):
            iri, label = result
            iri_str = iri.toPython()
            label_str = label.toPython()

            # if the label is also the URI then try to parse it.
            if label_str.startswith('http'):
                label_str = self._from_uri_to_text(label_str)

            label_str = re.sub('[^0-9a-zA-Z]+', ' ', label_str)

            ontology_relations[iri_str] = label_str

        return ontology_relations

    def _get_ontology_object_labels(self) -> dict[str, str]:
        """
        Returns a IRI (URI) to label dictionary.
        """
        results = self.graph.query(self.OBJECTS_QUERY)
        object_labels = {}
        for result in tqdm(results, desc='Loading Ontology Object Labels'):
            iri, label = result
            iri_str = iri.toPython()
            label_str = label.toPython()

            if iri_str == label_str and not label_str.startswith('http'):
                # skip BNodes
                continue

            # if the label is also the URI then try to parse it.
            if label_str.startswith('http'):
                label_str = self._from_uri_to_text(label_str)
            # Convert label to lower case snake case and remove spaces.
            label_str = self._camel_to_snake(label_str)
            label_str = re.sub('[^0-9a-zA-Z]+', ' ', label_str)

            object_labels[iri_str] = label_str

        return object_labels

    def _util_lookup(self, dictionary, val):
        """
        Helper function used to perform lookup in a given dictionary using a key.

        If the value is found, it is returned.
        If the value is in the ignore list, a `IGNORE_VALUE` object is returned.
        If a rephrase configuration is set of the key, it is returned.
        If not found, text extraction is attempt from the key URI.

        :param dictionary: The dictionary to look in.
        :param val: The key to lookup with.
        :return: The value.
        """
        if isinstance(val, URIRef):
            val = val.toPython()

        if not isinstance(val, str):
            return None

        if val in self._ignore_list:
            return self.__class__.IGNORE_VALUE

        # try get from overrides first
        if result := self.rephrased.get(val):
            return result

        result = dictionary.get(val)
        if result:
            return result

        # result is none, try to verbalize from URI
        return self._from_uri_to_text(val)

    def _from_uri_to_text(self, uri: str):
        """
        Attempt to extract phrases from the URI by taking the suffix and converting it into snake-case.
        """
        if '#' in uri:
            text = uri.split('#')[1]
        else:
            text = uri.split('/')[-1]

        # convert underscore or camel case notation into regular text
        text = self._camel_to_snake(text)
        return text.replace('_', ' ')

    @staticmethod
    def _camel_to_snake(name):
        """
        Convert camelCase into snake_case.
        e.g. MyPhrase -> my_phrase
        """
        name = re.sub('(.)([A-Z][a-z]+)', r'\1_\2', name)
        return re.sub('([a-z0-9])([A-Z])', r'\1_\2', name).lower()

    def _load_imports(self):
        """
        Loads concepts from imports. This is done by creating a new instance of Vocabulary, using the imported graph.
        """
        owl_imports = {o[0].toPython() for o in self.graph.query("SELECT DISTINCT ?o WHERE { ?s owl:imports ?o }")}
        for owl_import in owl_imports:
            logging.info(f'LOADING IMPORT: {owl_import}')
            graph = Graph()
            graph.parse(owl_import, format='xml')
            sub_vocab = self.__class__(graph)
            self.object_labels.update(sub_vocab.object_labels)
            self.relationship_labels.update(sub_vocab.relationship_labels)



import datetime
import logging
from pathlib import Path
from typing import Optional
from xml.sax import SAXParseException

import pandas
from rdflib import Graph, URIRef
from tqdm import tqdm

from verbalizer.nlp import ParaphraseLanguageModel
from verbalizer.sampler import Sampler
from verbalizer.verbalizer import Verbalizer

logger = logging.getLogger(__name__)


class Processor:
    """
    The processor that starts the verbalization process and outputs the results.
    """

    @classmethod
    def verbalize_with(cls,
                       verbalizer: Verbalizer,
                       *,
                       namespace: str,
                       output_dir: Optional[str] = None,
                       chunk_size: int = 1000,
                       sampler: Optional[Sampler] = None,
                       as_generator: bool = False):
        gen = cls.verbalize_with_stream(
            verbalizer,
            namespace=namespace,
            output_dir=output_dir,
            chunk_size=chunk_size,
            sampler=sampler,
            as_generator=as_generator
        )
        if as_generator:
            return gen

        return next(gen)

    @classmethod
    def verbalize_with_stream(
            cls,
            verbalizer: Verbalizer,
            *,
            namespace: str,
            output_dir: Optional[str] = None,
            chunk_size: int = 1000,
            sampler: Optional[Sampler] = None,
            as_generator: bool = False):
        """
        Start the verbalization process.
        :param verbalizer: The verbalizer to use.
        :param namespace: Name of the directory to create under the output directory.
        :param output_dir: Name of the output directory.
        :param chunk_size: Number of entries (rows) per file. default = 1000
        :param sampler: A sampling configuration, use to sample large ontologies.
        :param as_generator: If True, returns a generator instead of a list.
        """

        # current timestamp
        now = datetime.datetime.now(datetime.UTC)
        timestamp = int(now.timestamp())
        llm: ParaphraseLanguageModel = verbalizer.llm

        # make output directory
        if output_dir:
            if llm:
                out = f'{output_dir}/{namespace}/{llm.name}/{timestamp}'
            else:
                out = f'{output_dir}/{namespace}/{timestamp}'
        else:
            out = None

        classes = cls._get_classes(verbalizer.graph)
        individuals = cls._get_individuals(verbalizer.graph)

        if sampler := sampler:
            samples = sampler.get_sample({'classes': classes, 'individuals': individuals})
            classes, individuals = samples['classes'], samples['individuals']
        full_dataset = []
        chunk_dataset = []

        partition = 0
        for entry in tqdm(classes + individuals, desc='Verbalizing'):
            fragment, text, llm_text, stats = verbalizer.verbalize(entry)

            if stats.statements == 0:
                continue

            element = {
                'ontology': namespace,
                'root': entry,
                'fragment': fragment,
                'text': text,
                'llm_text': llm_text,
                'model': llm.name if llm else 'None',
                'statements': stats.statements,
                'unique_concepts': len(stats.concepts),
                'unique_relationships': len(stats.relationship_counter),
                'total_relationships': sum(stats.relationship_counter.values()),
                **stats.relationship_counter
            }

            chunk_dataset.append(element)

            if as_generator:
                yield element

            if len(chunk_dataset) != chunk_size:
                continue

            full_dataset.extend(chunk_dataset)

            if out:
                Path(out).mkdir(parents=True, exist_ok=True)
                pandas.DataFrame(chunk_dataset).to_csv(f'{out}/file_{partition}.csv', index=False)

            partition += 1
            chunk_dataset = []

        # handle leftovers
        if chunk_dataset:
            full_dataset.extend(chunk_dataset)
            if out:
                Path(out).mkdir(parents=True, exist_ok=True)
                pandas.DataFrame(chunk_dataset).to_csv(f'{out}/file_{partition}.csv', index=False)

        logger.info('Finished verbalizing')
        if llm:
            logger.info(f'LLM usage cost: ${llm.cost}')

        if not as_generator:
            yield full_dataset

    @staticmethod
    def _get_classes(graph):
        """
        Get all owl:Class and their labels.
        :param graph: The ontology.
        :return: A list of URIRef objects.
        """
        query = """
                SELECT ?o ?label
                WHERE {
                    ?o a owl:Class .
                    OPTIONAL {
                        ?o rdfs:label ?label
                    }
                    FILTER NOT EXISTS {
                        ?o owl:deprecated true
                    }
                }
            """
        return [result[0] for result in
                tqdm(graph.query(query), desc='Loading Classes') if isinstance(result[0], URIRef)]

    @staticmethod
    def _get_individuals(graph):
        """
        Get all owl:NamedIndividual and their labels.
        :param graph: The ontology.
        :return: A list of URIRef objects.
        """
        query = """
                        SELECT ?o ?label
                        WHERE {
                            ?o a owl:NamedIndividual .
                            OPTIONAL {
                                ?o rdfs:label ?label
                            }
                            FILTER NOT EXISTS {
                                ?o owl:deprecated true
                            }
                        }
                    """
        return [result[0] for result in
                tqdm(graph.query(query), desc='Loading Instances') if isinstance(result[0], URIRef)]

    @staticmethod
    def from_file(file_path: str) -> Graph:
        """
        Helper function to load graph from file.
        """
        graph = Graph()
        formats = ['xml', 'n3']
        logger.info(f'Loading File {file_path}')
        for file_format in formats:
            try:
                graph.parse(file_path, format=file_format)
                break
            except SAXParseException:
                pass
        logger.info(f'Done Loading.')
        return graph



from abc import ABC, abstractmethod

from rdflib import Graph
from rdflib.term import Node

from verbalizer.vocabulary import Vocabulary


class Pattern(ABC):
    """
    A patterns is a class used to identify patterns within the graph based on all the out going relations from a
    specific concept.
    """

    def __init__(self, graph: Graph, verbalizer: 'Verbalizer', vocabulary: Vocabulary):
        self._graph = graph
        self.verbalizer = verbalizer
        self.vocab = vocabulary

    @abstractmethod
    def check(self, results) -> bool:
        """
        The check function is used to determine whether a patterns was detected or not. The "results" argument
        includes the first-degree related objects to the subject. If more triples are needed to be fetched to
        identify the patterns, self.graph.query can be used.
        """
        return False

    @abstractmethod
    def normalize(self, node: 'VerbalizationNode', triple_collector) -> list[tuple[Node, Node]]:
        """
        The normalize function is called only if the check returned True. It is used to simplify the patterns. This
        means that the patterns must re-arrange the connected nodes in an order that would be different from the
        regular order. In addition to that, the implementation must also collect all the RDF triples observed
        regardless of whether they were used or not in the construction of the nodes and edges.
        """
        return []
